package net.lemnik.eodsql.spi.util;

import java.lang.reflect.Type;
import java.lang.reflect.ParameterizedType;

import java.sql.ResultSet;
import java.sql.SQLException;

import java.util.Map;
import java.util.Set;
import java.util.List;
import java.util.TreeSet;
import java.util.HashSet;
import java.util.SortedSet;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;

import net.lemnik.eodsql.InvalidDataTypeException;

import net.lemnik.eodsql.spi.util.DataObjectBinding.BindingType;

/**
 * Created on 2008/06/29
 * @author Jason Morris
 */
class CollectionWrapper extends AbstractResultSetWrapper<Collection, Object> {

    private final Class<? extends Collection> typeClass;

    private CollectionWrapper(
            final Class<? extends Collection> typeClass,
            final DataObjectBinding<Object> binding) {

        super(binding);
        this.typeClass = typeClass;
    }

    @Override
    public Collection wrap(final ResultSet results) throws SQLException {
        /*
         * An important note on the Collection wrapper: this class sticks
         * to using standard Java Collections. The main reason for this is
         * that it makes it very easy for outside Frameworks (ex: GWT) to
         * bind these implementations without supporting the entire Java
         * Serialization mechanism.
         */

        if(typeClass == List.class || typeClass == Collection.class) {
            final ArrayList<Object> output = new ArrayList<Object>();
            wrap(results, output);
            output.trimToSize();

            return Collections.unmodifiableList(output);
        } else if(typeClass == Set.class) {
            final HashSet<Object> output = new HashSet<Object>();
            wrap(results, output);

            return Collections.unmodifiableSet(output);
        } else if(typeClass == SortedSet.class) {
            final TreeSet<Object> output = new TreeSet<Object>();
            wrap(results, output);

            return Collections.unmodifiableSortedSet(output);
        }

        return null;
    }

    private void wrap(
            final ResultSet results,
            final Collection<Object> in)
            throws SQLException {

        while(results.next()) {
            in.add(binding.unmarshall(results));
        }
    }

    @Override
    public int getPreferredResultSetConcurrency() {
        return ResultSet.CONCUR_READ_ONLY;
    }

    @Override
    public int getPreferredResultSetType() {
        return ResultSet.TYPE_FORWARD_ONLY;
    }

    static final class Factory implements ResultSetWrapper.Factory {

        public boolean isTypeConstructable(
                final Type type,
                final Map<String, Object> parameters)
                throws InvalidDataTypeException {

            if(type instanceof ParameterizedType) {
                final ParameterizedType paramType = (ParameterizedType)type;

                if(paramType.getRawType() instanceof Class &&
                        paramType.getActualTypeArguments().length == 1) {

                    final Class clazz = (Class)paramType.getRawType();
                    final Class dataObject = (Class)paramType.
                            getActualTypeArguments()[0];

                    if(clazz == SortedSet.class) {
                        if(Comparable.class.isAssignableFrom(dataObject)) {
                            return true;
                        } else {
                            throw new InvalidDataTypeException(
                                    "DataType " + dataObject.getName() +
                                    " cannot be returned in a " +
                                    clazz.getName() + " because" +
                                    " it does not implement Comparable.",
                                    dataObject);
                        }
                    }

                    return clazz == Set.class ||
                            clazz == List.class ||
                            clazz == Collection.class;
                }
            }

            return false;
        }

        @SuppressWarnings("unchecked")
        public ResultSetWrapper create(
                final Type genericType,
                final Map<String, Object> parameters) {

            final ParameterizedType paramType = (ParameterizedType)genericType;
            final BindingType bindingType =
                    parameters.containsKey(PARAMETER_BINDING_TYPE)
                    ? (BindingType)parameters.get(PARAMETER_BINDING_TYPE)
                    : BindingType.NORMAL_BINDING;

            final Class<?> dataObject = (Class<?>)
                    paramType.getActualTypeArguments()[0];

            final DataObjectBinding binding;
            if (parameters.containsKey(PARAMETER_CUSTOM_DATA_OBJECT_BINDING)) {
                final DataObjectBinding<Object> 
                    customBinding = (DataObjectBinding<Object>) parameters.get(PARAMETER_CUSTOM_DATA_OBJECT_BINDING);
                customBinding.setObjectType((Class<Object>) dataObject);
                binding = customBinding;
            } else {
                binding = DataObjectBinding.getDataObjectBinding(dataObject, bindingType);
            }

            return new CollectionWrapper(
                    (Class<? extends Collection>)paramType.getRawType(),
                    binding);
        }

    }
}
